#include "backend/passes.hpp"
#include <iostream>
#include <memory>
#include <vector>
#include <set>

namespace backend{
namespace riscv{

using namespace RiscvInstr;
using namespace RiscvReg;

static Instr* optimize_constant_reg_inst(Function *func, Instr *inst) {
    if(auto binary_inst = dynamic_cast<Binary*>(inst)){
        if(func->constant_regs.find(binary_inst->use_reg()[0]) != func->constant_regs.end() && func->constant_regs.find(binary_inst->use_reg()[1]) != func->constant_regs.end()){
            // lhs and rhs is constant, can directly calc value
            int32_t result;
            switch (binary_inst->get_op()) {
                case RvBinaryOp::ADD:
                    result = func->constant_regs[binary_inst->use_reg()[0]] + func->constant_regs[binary_inst->use_reg()[1]];
                    func->constant_regs[binary_inst->def_reg()[0]] = result;
                    return new LoadImm(binary_inst->def_reg()[0], result);
                case RvBinaryOp::SUB:
                    result = func->constant_regs[binary_inst->use_reg()[0]] - func->constant_regs[binary_inst->use_reg()[1]];
                    func->constant_regs[binary_inst->def_reg()[0]] = result;
                    return new LoadImm(binary_inst->def_reg()[0], result);
                case RvBinaryOp::MUL:
                    result = func->constant_regs[binary_inst->use_reg()[0]] * func->constant_regs[binary_inst->use_reg()[1]];
                    func->constant_regs[binary_inst->def_reg()[0]] = result;
                    return new LoadImm(binary_inst->def_reg()[0], result);
                case RvBinaryOp::DIV:
                    result = func->constant_regs[binary_inst->use_reg()[0]] / func->constant_regs[binary_inst->use_reg()[1]];
                    func->constant_regs[binary_inst->def_reg()[0]] = result;
                    return new LoadImm(binary_inst->def_reg()[0], result);
                case RvBinaryOp::REM:
                    result = func->constant_regs[binary_inst->use_reg()[0]] % func->constant_regs[binary_inst->use_reg()[1]];
                    func->constant_regs[binary_inst->def_reg()[0]] = result;
                    return new LoadImm(binary_inst->def_reg()[0], result);
                case RvBinaryOp::OR:
                    result = func->constant_regs[binary_inst->use_reg()[0]] | func->constant_regs[binary_inst->use_reg()[1]];
                    func->constant_regs[binary_inst->def_reg()[0]] = result;
                    return new LoadImm(binary_inst->def_reg()[0], result);
                case RvBinaryOp::AND:
                    result = func->constant_regs[binary_inst->use_reg()[0]] & func->constant_regs[binary_inst->use_reg()[1]];
                    func->constant_regs[binary_inst->def_reg()[0]] = result;
                    return new LoadImm(binary_inst->def_reg()[0], result);
                case RvBinaryOp::XOR:
                    result = func->constant_regs[binary_inst->use_reg()[0]] ^ func->constant_regs[binary_inst->use_reg()[1]];
                    func->constant_regs[binary_inst->def_reg()[0]] = result;
                    return new LoadImm(binary_inst->def_reg()[0], result);
                case RvBinaryOp::SLT:
                    result = func->constant_regs[binary_inst->use_reg()[0]] < func->constant_regs[binary_inst->use_reg()[1]];
                    func->constant_regs[binary_inst->def_reg()[0]] = result;
                    return new LoadImm(binary_inst->def_reg()[0], result);
                case RvBinaryOp::SLTU:
                    result = (unsigned)func->constant_regs[binary_inst->use_reg()[0]] < (unsigned)func->constant_regs[binary_inst->use_reg()[1]];
                    func->constant_regs[binary_inst->def_reg()[0]] = result;
                    return new LoadImm(binary_inst->def_reg()[0], result);
                case RvBinaryOp::SLL:
                    result = func->constant_regs[binary_inst->use_reg()[0]] << func->constant_regs[binary_inst->use_reg()[1]];
                    func->constant_regs[binary_inst->def_reg()[0]] = result;
                    return new LoadImm(binary_inst->def_reg()[0], result);
                case RvBinaryOp::SRL:
                    result = func->constant_regs[binary_inst->use_reg()[0]] >> func->constant_regs[binary_inst->use_reg()[1]];
                    func->constant_regs[binary_inst->def_reg()[0]] = result;
                    return new LoadImm(binary_inst->def_reg()[0], result);
                default:
                    break;
            }
        }
        if(func->constant_regs.find(binary_inst->use_reg()[0]) != func->constant_regs.end()){
            // lhs is constant
            switch (binary_inst->get_op()) {
                case RvBinaryOp::ADD:
                    return new ADDImm(binary_inst->def_reg()[0], binary_inst->use_reg()[1], func->constant_regs[binary_inst->use_reg()[0]]);
                case RvBinaryOp::SUB:
                    return new ADDImm(binary_inst->def_reg()[0], binary_inst->use_reg()[1], -func->constant_regs[binary_inst->use_reg()[0]]);
                default:
                    break;
            }
        }
        if(func->constant_regs.find(binary_inst->use_reg()[1]) != func->constant_regs.end()){
            // lhs is constant
            switch (binary_inst->get_op()) {
                case RvBinaryOp::ADD:
                    return new ADDImm(binary_inst->def_reg()[0], binary_inst->use_reg()[0], func->constant_regs[binary_inst->use_reg()[1]]);
                case RvBinaryOp::SUB:
                    return new ADDImm(binary_inst->def_reg()[0], binary_inst->use_reg()[0], -func->constant_regs[binary_inst->use_reg()[1]]);
                default:
                    break;
            }
        }
        
    }
    return nullptr;
}
void inline_constant(Function *func){
    for(auto bb: *func->get_bbs()){
        for(auto inst: *bb->get_instr()){
            auto replace = optimize_constant_reg_inst(func, inst);
            if(replace){
                inst = std::move(replace);
            }
        }
    }
}

void optimize_before_reg_alloc(Program *prog){
    for(auto func: prog->functions()) {
        for (auto it = func->constant_regs.begin(); it != func->constant_regs.end(); it++)
            std::cout << it->first << ": " << it->second << std::endl;
        inline_constant(func);
    }
}




}
}