#pragma once

#include "middleend/ir.hpp"
#include "middleend/temp.hpp"

namespace middleend {

class CFG {
private:
    ir::Function *func;
    std::list<int> bb_list;
    std::unordered_set<int> jump_in_bb;
    std::unordered_map<int, ir::BasicBlock *> bb_map;
    std::unordered_map<int, std::unordered_set<int>> bb_prev;
    std::unordered_map<int, std::unordered_set<int>> bb_succ;
    void add_edge(int from, int to);
    void build_bb_edge();
    void remove_unreachable_bb(); // proceed in build_bb_edge()
public:
    CFG(ir::Function *func);
    ir::Function* get_func() { return func; }
    inline std::list<int>* get_bb_list() { return &bb_list; }
    inline std::unordered_map<int, ir::BasicBlock *>* get_bb_map() { return &bb_map; }
    inline std::unordered_map<int, std::unordered_set<int>>* get_bb_prev_map() { return &bb_prev; }
    inline std::unordered_map<int, std::unordered_set<int>>* get_bb_succ_map() { return &bb_succ; }
    inline ir::BasicBlock* get_bb(int bb) { return bb_map[bb]; }
    inline std::unordered_set<int>* get_bb_prev(int bb) { return &bb_prev[bb]; }
    inline std::unordered_set<int>* get_bb_succ(int bb) { return &bb_succ[bb]; }
    void print();
    void rebuild();
};

} // namespace middleend