#include "middleend/remove_useless_inst.hpp"
#include <iostream>

namespace middleend {

bool RemoveUselessInst::remove_useless_definition() {
    bool flag_ret = false;
    std::unordered_set<int> used_temp;
    for (auto bb : *func_->get_basic_blocks()) {
        for (auto inst : *bb->get_instructions()) {
            if (auto store = dynamic_cast<ir::instruction::Store*>(inst)) {
                // std::cout << "store: " << store->to_str() << std::endl;
                used_temp.insert(store->getaddr()->get_index());
                for (auto src : *inst->get_src()) {
                    used_temp.insert(src->get_index());
                }
            } else {
                // std::cout << "others: " << inst->to_str() << std::endl;
                for (auto src : *inst->get_src()) {
                    if(src == nullptr) continue;
                    used_temp.insert(src->get_index());
                }
            }
        }
    }
    for (auto bb : *func_->get_basic_blocks()) {
        for (std::vector<ir::Instruction *>::iterator ins = bb->get_instructions()->begin(); ins != bb->get_instructions()->end(); ins++) {
            if ((*ins)->get_instr_kind() != ir::IRInstrKind::IRInstr_SEQ) continue;
            // if (auto load = dynamic_cast<ir::instruction::Load*>(*ins)) continue;
            // if (auto store = dynamic_cast<ir::instruction::Store*>(*ins)) continue;
            bool flag = false;
            for (auto dst : *(*ins)->get_dst()) {
                if (used_temp.count(dst->get_index())) {
                    flag = true;
                    break;
                }
            }
            if (!flag) {
                bb->get_instructions()->erase(ins);
                ins--;
                flag_ret = true;
            }
        }
    }
    return flag_ret;
}

bool RemoveUselessInst::remove_useless_jump() {
    bool flag_ret = false;
    int bb_count = func_->get_basic_blocks()->size();
    for (int i = 0; i < bb_count - 1; i++) {
        ir::BasicBlock *bb = func_->get_basic_blocks()->at(i);
        ir::BasicBlock *next = func_->get_basic_blocks()->at(i + 1);
        if (bb->get_instructions()->size() == 0) continue;
        auto last = bb->get_instructions()->back();
        if (last->get_instr_kind() == ir::IRInstrKind::IRInstr_JMP) {
            auto jump = dynamic_cast<ir::instruction::Branch *>(last);
            if (jump->get_bb() == next) {
                bb->get_instructions()->pop_back();
                flag_ret = true;
            }
        }
    }
    return flag_ret;
}

// void RemoveUselessInst::rebuild_bb_edge() {
//     for (auto bb : *func_->get_basic_blocks()) {
//         bb->get_succ()->clear();
//         bb->get_prev()->clear();
//     }
//     for (auto bb : *func_->get_basic_blocks()) {
//         if (bb->get_instructions()->size() == 0) continue;
//         auto last = bb->get_instructions()->back();
//         if (last->get_instr_kind() == ir::IRInstrKind::IRInstr_JMP) {
//             auto branch = dynamic_cast<ir::instruction::Branch *>(last);
//             bb->get_succ()->insert(branch->get_bb());
//             branch->get_bb()->get_prev()->insert(bb);
//         } else if (last->get_instr_kind() == ir::IRInstrKind::IRInstr_JMP_COND) {
//             auto condbranch = dynamic_cast<ir::instruction::CondBranch *>(last);
//             bb->get_succ()->insert(condbranch->get_true_bb());
//             bb->get_succ()->insert(condbranch->get_false_bb());
//             condbranch->get_true_bb()->get_prev()->insert(bb);
//             condbranch->get_false_bb()->get_prev()->insert(bb);
//         } else if (bb->get_index() + 1 < func_->get_basic_blocks()->size()) {
//             bb->get_succ()->insert(func_->get_basic_blocks()->at(bb->get_index() + 1));
//             func_->get_basic_blocks()->at(bb->get_index() + 1)->get_prev()->insert(bb);
//         }
//     }
// }

// bool RemoveUselessInst::rebuild_bb() {
//     bool flag_ret = false;
//     for (auto bb_iter = func_->get_basic_blocks()->begin(); bb_iter != func_->get_basic_blocks()->end(); bb_iter++) {
//         if ((*bb_iter)->get_instructions()->size() == 0) {
//             func_->get_basic_blocks()->erase(bb_iter);
//             bb_iter--;
//             flag_ret = true;
//         }
//     }
//     rebuild_bb_edge();
//     for (auto bb : *func_->get_basic_blocks()) bb->set_visit(false);
//     std::list<ir::BasicBlock *> q;
//     q.push_back(func_->get_basic_blocks()->front());
//     func_->get_basic_blocks()->front()->set_visit(true);
//     while (!q.empty()) {
//         ir::BasicBlock *bb = q.front();
//         q.pop_front();
//         for (auto succ : *bb->get_succ()) {
//             if (!succ->get_visit()) {
//                 succ->set_visit(true);
//                 q.push_back(succ);
//             }
//         }
//     }
//     std::unordered_set<int> removed_bb;
//     for (auto bb : *func_->get_basic_blocks()) if (!bb->get_visit()) {
//         removed_bb.insert(bb->get_index());
//         for (auto prev : *bb->get_prev()) prev->get_succ()->erase(bb);
//         for (auto succ : *bb->get_succ()) succ->get_prev()->erase(bb);
//     }
//     for (auto bb_iter = func_->get_basic_blocks()->begin(); bb_iter != func_->get_basic_blocks()->end(); bb_iter++) {
//         if (!(*bb_iter)->get_visit()) {
//             func_->get_basic_blocks()->erase(bb_iter);
//             bb_iter--;
//             flag_ret = true;
//         }
//     }
//     for (auto bb : *func_->get_basic_blocks()) {
//         for (auto inst : *bb->get_instructions()) {
//             if (auto phi = dynamic_cast<ir::instruction::Phi *>(inst)) {
//                 int new_count = 0;
//                 for (int i = 0; i < phi->getbbs()->size(); i++) {
//                     if (!removed_bb.count(phi->getbbs()->at(i)->get_index())) {
//                         phi->getbbs()->at(new_count) = phi->getbbs()->at(i);
//                         phi->getvalues()->at(new_count) = phi->get_src()->at(i);
//                         new_count++;
//                     }
//                 }
//                 phi->getbbs()->resize(new_count);
//                 phi->getvalues()->resize(new_count);
//             }
//         }
//     }
//     return flag_ret;
// }

} // namespace middleend